@@ -10,7 +10,7 @@
 #include "biom.hpp"
 #include "unifrac.hpp"
 
-enum Format {format_invalid,format_ascii, format_hdf5_fp32, format_hdf5_fp64};
+enum Format {format_invalid,format_ascii, format_hdf5_fp32, format_hdf5_fp64, format_hdf5_nodist};
 
 void usage() {
     std::cout << "usage: ssu -i <biom> -o <out.dm> -m [METHOD] -t <newick> [-a alpha] [-f]  [--vaw]" << std::endl;
@@ -24,6 +24,8 @@ void usage() {
     std::cout << "                       unweighted_fp64 | weighted_normalized_fp64 | weighted_unnormalized_fp64 | generalized_fp64 |" << std::endl;
     std::cout << "                       unweighted_fp32 | weighted_normalized_fp32 | weighted_unnormalized_fp32 | generalized_fp32]." << std::endl;
     std::cout << "    -o\t\tThe output distance matrix." << std::endl;
+    std::cout << "    -g\t\t[OPTIONAL] The input grouping in TSV." << std::endl;
+    std::cout << "    -c\t\t[OPTIONAL] The columns(s) to use for grouping, multiple values comma separated." << std::endl;
     std::cout << "    -a\t\t[OPTIONAL] Generalized UniFrac alpha, default is 1." << std::endl;
     std::cout << "    -f\t\t[OPTIONAL] Bypass tips, reduces compute by about 50%." << std::endl;
     std::cout << "    --vaw\t[OPTIONAL] Variance adjusted, default is to not adjust for variance." << std::endl;
@@ -33,6 +35,7 @@ void usage() {
     std::cout << "    \t\t    partial-report : Start and stop suggestions for partial compute." << std::endl;
     std::cout << "    \t\t    merge-partial : Merge partial UniFrac results." << std::endl;
     std::cout << "    \t\t    check-partial : Check partial UniFrac results." << std::endl;
+    std::cout << "    \t\t    multi : compute UniFrac multiple times." << std::endl;
     std::cout << "    --start\t[OPTIONAL] If mode==partial, the starting stripe." << std::endl;
     std::cout << "    --stop\t[OPTIONAL] If mode==partial, the stopping stripe." << std::endl;
     std::cout << "    --partial-pattern\t[OPTIONAL] If mode==merge-partial or check-partial, a glob pattern for partial outputs to merge." << std::endl;
@@ -40,11 +43,16 @@ void usage() {
     std::cout << "    --report-bare\t[OPTIONAL] If mode==partial-report, produce barebones output." << std::endl;
     std::cout << "    --n-substeps\t[OPTIONAL] Internally split the problem in n substeps for reduced memory footprint, default is 1." << std::endl;
     std::cout << "    --format|-r\t[OPTIONAL]  Output format:" << std::endl;
-    std::cout << "    \t\t    ascii : [DEFAULT] Original ASCII format." << std::endl;
-    std::cout << "    \t\t    hfd5 : HFD5 format.  May be fp32 or fp64, depending on method." << std::endl;
+    std::cout << "    \t\t    ascii : Original ASCII format. (default if mode==one-off)" << std::endl;
+    std::cout << "    \t\t    hdf5 : HFD5 format.  May be fp32 or fp64, depending on method." << std::endl;
     std::cout << "    \t\t    hdf5_fp32 : HFD5 format, using fp32 precision." << std::endl;
     std::cout << "    \t\t    hdf5_fp64 : HFD5 format, using fp64 precision." << std::endl;
+    std::cout << "    \t\t    hdf5_nodist : HFD5 format, no distance matrix. (default if mode==multi)" << std::endl;
+    std::cout << "    --subsample-depth\tDepth of subsampling of the input BIOM before computing unifrac (required for mode==multi, optional for one-off)" << std::endl;
+    std::cout << "    --n-subsamples\t[OPTIONAL] if mode==multi, number of subsampled UniFracs to compute (default: 100)" << std::endl;
+    std::cout << "    --permanova\t[OPTIONAL] Number of PERMANOVA permutations to compute (default: 999 with -g, do not compute if 0)" << std::endl;
     std::cout << "    --pcoa\t[OPTIONAL] Number of PCoA dimensions to compute (default: 10, do not compute if 0)" << std::endl;
+    std::cout << "    --seed\t[OPTIONAL] Seed to use for initializing the random gnerator" << std::endl;
     std::cout << "    --diskbuf\t[OPTIONAL] Use a disk buffer to reduce memory footprint. Provide path to a fast partition (ideally NVMe)." << std::endl;
     std::cout << "    -n\t\t[OPTIONAL] DEPRECATED, no-op." << std::endl;
     std::cout << std::endl;
@@ -76,13 +84,15 @@ void usage() {
     std::cout << std::endl;
 }
 
-const char* compute_status_messages[7] = {"No error.",
+const char* compute_status_messages[9] = {"No error.",
                                           "The tree file cannot be found.", 
                                           "The table file cannot be found.",
                                           "The table file contains an empty table.",
                                           "An unknown method was requested.", 
                                           "Table observation IDs are not a subset of the tree tips. This error can also be triggered if a node name contains a single quote (this is unlikely).",
-                                          "Error creating the output."};
+                                          "Error creating the output.",
+                                          "The requested method is not supported.",
+                                          "The grouping file cannot be found or does not have the necessary data."};
 
 
 // https://stackoverflow.com/questions/8401777/simple-glob-in-c-on-unix-system
@@ -115,7 +125,7 @@ int mode_partial_report(const std::string table_filename, unsigned int npartials
         exit(EXIT_FAILURE);
     }
 
-    su::biom table = su::biom(table_filename.c_str());
+    su::biom table(table_filename.c_str());
     int total_stripes = (table.n_samples + 1) / 2;
 
     if(!bare) {
@@ -154,6 +164,7 @@ int mode_partial_report(const std::string table_filename, unsigned int npartials
 } 
 
 int mode_merge_partial_fp32(const char * output_filename, Format format_val, unsigned int pcoa_dims,
+                            unsigned int permanova_perms, const char *grouping_filename, const char *grouping_columns,
                             size_t partials_size, partial_dyn_mat_t* * partial_mats,
                             const char * mmap_dir) {
     mat_full_fp32_t *result = NULL;
@@ -167,8 +178,10 @@ int mode_merge_partial_fp32(const char * output_filename, Format format_val, uns
         return EXIT_FAILURE;
     }
 
+    // TODO: Add support for PERMANOVA
+
     IOStatus iostatus;
-    iostatus = write_mat_from_matrix_hdf5_fp32(output_filename, result, pcoa_dims);
+    iostatus = write_mat_from_matrix_hdf5_fp32(output_filename, result, pcoa_dims, format_val!=format_hdf5_nodist);
     destroy_mat_full_fp32(&result);
     
     if(iostatus != write_okay) {
@@ -182,6 +195,7 @@ int mode_merge_partial_fp32(const char * output_filename, Format format_val, uns
 }
 
 int mode_merge_partial_fp64(const char * output_filename, Format format_val, unsigned int pcoa_dims,
+                            unsigned int permanova_perms, const char *grouping_filename, const char *grouping_columns,
                             size_t partials_size, partial_dyn_mat_t* * partial_mats,
                             const char * mmap_dir) {
     mat_full_fp64_t *result = NULL;
@@ -195,9 +209,11 @@ int mode_merge_partial_fp64(const char * output_filename, Format format_val, uns
         return EXIT_FAILURE;
     }
 
+    // TODO: Add support for PERMANOVA
+
     IOStatus iostatus;
-    if (format_val==format_hdf5_fp64) {
-     iostatus = write_mat_from_matrix_hdf5(output_filename, result, pcoa_dims);
+    if (format_val!=format_ascii) {
+     iostatus = write_mat_from_matrix_hdf5_fp64(output_filename, result, pcoa_dims, format_val!=format_hdf5_nodist);
     } else {
      iostatus = write_mat_from_matrix(output_filename, result);
     }
@@ -215,6 +231,7 @@ int mode_merge_partial_fp64(const char * output_filename, Format format_val, uns
 
 
 int mode_merge_partial(const std::string &output_filename, Format format_val, unsigned int pcoa_dims,
+                       unsigned int permanova_perms, const std::string &grouping_filename, const std::string &grouping_columns,
                        const std::string &partial_pattern,
                        const std::string &mmap_dir) {
     if(output_filename.empty()) {
@@ -230,6 +247,16 @@ int mode_merge_partial(const std::string &output_filename, Format format_val, un
         return EXIT_FAILURE;
     }
     
+    if((permanova_perms>0) && grouping_filename.empty()) {
+        err("grouping filename missing");
+        return EXIT_FAILURE;
+    }
+    
+    if((permanova_perms>0) && grouping_columns.empty()) {
+        err("grouping columns missing");
+        return EXIT_FAILURE;
+    }
+    
     std::vector<std::string> partials = glob(partial_pattern);
     partial_dyn_mat_t** partial_mats = (partial_dyn_mat_t**)malloc(sizeof(partial_dyn_mat_t*) * partials.size());
     for(size_t i = 0; i < partials.size(); i++) {
@@ -243,14 +270,22 @@ int mode_merge_partial(const std::string &output_filename, Format format_val, un
     }
 
     const char * mmap_dir_c = mmap_dir.empty() ? NULL : mmap_dir.c_str();
+    const char * grouping_c = (permanova_perms>0) ? grouping_filename.c_str() : NULL ;
+    const char * columns_c = (permanova_perms>0) ? grouping_columns.c_str() : NULL ;
 
     int status;
     if (format_val==format_hdf5_fp64) {
-     status = mode_merge_partial_fp64(output_filename.c_str(), format_val, pcoa_dims, partials.size(), partial_mats, mmap_dir_c);
+     status = mode_merge_partial_fp64(output_filename.c_str(), format_val,
+                                      pcoa_dims, permanova_perms, grouping_c, columns_c,
+                                      partials.size(), partial_mats, mmap_dir_c);
     } else if (format_val==format_hdf5_fp32) {
-     status = mode_merge_partial_fp32(output_filename.c_str(), format_val, pcoa_dims, partials.size(), partial_mats, mmap_dir_c);
+     status = mode_merge_partial_fp32(output_filename.c_str(), format_val,
+                                      pcoa_dims, permanova_perms, grouping_c, columns_c,
+                                      partials.size(), partial_mats, mmap_dir_c);
     } else {
-     status = mode_merge_partial_fp64(output_filename.c_str(), format_val, pcoa_dims, partials.size(), partial_mats, mmap_dir_c);
+     status = mode_merge_partial_fp64(output_filename.c_str(), format_val,
+                                      pcoa_dims, permanova_perms, grouping_c, columns_c,
+                                      partials.size(), partial_mats, mmap_dir_c);
     }
 
     for(size_t i = 0; i < partials.size(); i++) {
@@ -358,7 +393,8 @@ int mode_partial(std::string table_filename, std::string tree_filename,
 
 int mode_one_off(const std::string &table_filename, const std::string &tree_filename, 
                  const std::string &output_filename, const std::string &format_str, Format format_val, 
-                 const std::string &method_string, unsigned int pcoa_dims,
+                 const std::string &method_string, unsigned int subsample_depth, unsigned int pcoa_dims,
+                 unsigned int permanova_perms, const std::string &grouping_filename, const std::string &grouping_columns,
                  bool vaw, double g_unifrac_alpha, bool bypass_tips,
                  unsigned int nsubsteps, const std::string &mmap_dir) {
     if(output_filename.empty()) {
@@ -376,6 +412,16 @@ int mode_one_off(const std::string &table_filename, const std::string &tree_file
         return EXIT_FAILURE;
     }
     
+    if((permanova_perms>0) && grouping_filename.empty()) {
+        err("grouping filename missing");
+        return EXIT_FAILURE;
+    }
+    
+    if((permanova_perms>0) && grouping_columns.empty()) {
+        err("grouping columns missing");
+        return EXIT_FAILURE;
+    }
+    
     if(method_string.empty()) {
         err("method missing");
         return EXIT_FAILURE;
@@ -385,6 +431,10 @@ int mode_one_off(const std::string &table_filename, const std::string &tree_file
     if (format_val==format_ascii) {
       mat_t *result = NULL;
 
+      if(subsample_depth>0) {
+        err("Subsampling not supported with ASCII output.");
+        return EXIT_FAILURE;
+      }
       status = one_off(table_filename.c_str(), tree_filename.c_str(), method_string.c_str(), 
                        vaw, g_unifrac_alpha, bypass_tips, nsubsteps, &result);
       if(status != okay || result == NULL) {
@@ -402,10 +452,13 @@ int mode_one_off(const std::string &table_filename, const std::string &tree_file
 
     } else {
       const char * mmap_dir_c = mmap_dir.empty() ? NULL : mmap_dir.c_str();
+      const char * grouping_c = (permanova_perms>0) ? grouping_filename.c_str() : NULL ;
+      const char * columns_c = (permanova_perms>0) ? grouping_columns.c_str() : NULL ;
 
-      status = unifrac_to_file(table_filename.c_str(), tree_filename.c_str(), output_filename.c_str(),
-                               method_string.c_str(), vaw, g_unifrac_alpha, bypass_tips, nsubsteps, format_str.c_str(),
-                               pcoa_dims, mmap_dir_c);
+      status = unifrac_to_file_v2(table_filename.c_str(), tree_filename.c_str(), output_filename.c_str(),
+                                  method_string.c_str(), vaw, g_unifrac_alpha, bypass_tips, nsubsteps, format_str.c_str(),
+                                  subsample_depth, true,
+                                  pcoa_dims, permanova_perms, grouping_c, columns_c, mmap_dir_c);
 
       if (status != okay) {
         fprintf(stderr, "Compute failed in one_off: %s\n", compute_status_messages[status]);
@@ -415,22 +468,100 @@ int mode_one_off(const std::string &table_filename, const std::string &tree_file
     return (status==okay) ? EXIT_SUCCESS : EXIT_FAILURE;
 }
 
+int mode_multi(const std::string &table_filename, const std::string &tree_filename, 
+               const std::string &output_filename, const std::string &format_str, Format format_val, 
+               const std::string &method_string,
+               unsigned int n_subsamples, unsigned int subsample_depth,
+               unsigned int pcoa_dims,
+               unsigned int permanova_perms, const std::string &grouping_filename, const std::string &grouping_columns,
+               bool vaw, double g_unifrac_alpha, bool bypass_tips,
+               unsigned int nsubsteps, const std::string &mmap_dir) {
+    if(output_filename.empty()) {
+        err("output filename missing");
+        return EXIT_FAILURE;
+    }
+
+    if(table_filename.empty()) {
+        err("table filename missing");
+        return EXIT_FAILURE;
+    }
+
+    if(tree_filename.empty()) {
+        err("tree filename missing");
+        return EXIT_FAILURE;
+    }
+    
+    if((permanova_perms>0) && grouping_filename.empty()) {
+        err("grouping filename missing");
+        return EXIT_FAILURE;
+    }
+    
+    if((permanova_perms>0) && grouping_columns.empty()) {
+        err("grouping columns missing");
+        return EXIT_FAILURE;
+    }
+    
+    if(method_string.empty()) {
+        err("method missing");
+        return EXIT_FAILURE;
+    }
+
+    if(subsample_depth<1) {
+      err("subsample_depth cannot be 0.");
+      return EXIT_FAILURE;
+    }
+
+    if(n_subsamples<1) {
+      err("n_subsamples cannot be 0.");
+      return EXIT_FAILURE;
+    }
+
+    if (format_val==format_ascii) {
+      err("ASCII format not supported in multi mode");
+      return EXIT_FAILURE;
+    }
+
+    compute_status status = okay;
+    {
+      const char * mmap_dir_c = mmap_dir.empty() ? NULL : mmap_dir.c_str();
+      const char * grouping_c = (permanova_perms>0) ? grouping_filename.c_str() : NULL ;
+      const char * columns_c = (permanova_perms>0) ? grouping_columns.c_str() : NULL ;
+
+      status = unifrac_multi_to_file_v2(table_filename.c_str(), tree_filename.c_str(), output_filename.c_str(),
+                                        method_string.c_str(), vaw, g_unifrac_alpha, bypass_tips, nsubsteps, format_str.c_str(),
+                                        n_subsamples, subsample_depth, true,
+                                        pcoa_dims, permanova_perms, grouping_c, columns_c, mmap_dir_c);
+
+      if (status != okay) {
+        fprintf(stderr, "Compute failed in multi: %s\n", compute_status_messages[status]);
+      }
+    }
+
+    return (status==okay) ? EXIT_SUCCESS : EXIT_FAILURE;
+}
+
 void ssu_sig_handler(int signo) {
     if (signo == SIGUSR1) {
         printf("Status cannot be reported.\n");
     }
 }
 
-Format get_format(const std::string &format_string, const std::string &method_string) {
+Format get_format(const std::string &format_string, const std::string &method_string, const std::string &mode_string) {
     Format format_val = format_invalid;
     if (format_string.empty()) {
-        format_val = format_ascii;
+        if (mode_string!="multi") {
+          format_val = format_ascii;
+        } else {
+          format_val = format_hdf5_nodist;
+        }
     } else if (format_string == "ascii") {
         format_val = format_ascii;
     } else if (format_string == "hdf5_fp32") {
         format_val = format_hdf5_fp32;
     } else if (format_string == "hdf5_fp64") {
         format_val = format_hdf5_fp64;
+    } else if (format_string == "hdf5_nodist") {
+        format_val = format_hdf5_nodist;
     } else if (format_string == "hdf5") {
         if ((method_string=="unweighted_fp64") || (method_string=="weighted_normalized_fp64") || (method_string=="weighted_unnormalized_fp64") || (method_string=="generalized_fp64"))
            format_val = format_hdf5_fp64;
@@ -441,6 +572,19 @@ Format get_format(const std::string &format_string, const std::string &method_st
     return format_val;
 }
 
+std::string format2str(Format format_val) {
+  if (format_val==format_hdf5_nodist) {
+    return "hdf5_nodist";
+  } else if (format_val==format_hdf5_fp32) {
+    return "hdf5_fp32";
+  } else if (format_val==format_hdf5_fp64) {
+    return "hdf5_fp64";
+  } else if (format_val==format_ascii) {
+    return "ascii";
+  } 
+  return "invalid";
+}
+
 int main(int argc, char **argv){
     signal(SIGUSR1, ssu_sig_handler);
     InputParser input(argc, argv);
@@ -456,6 +600,8 @@ int main(int argc, char **argv){
     std::string method_string = input.getCmdOption("-m");
     // deprecated, but we still want to support it, even as a no-op
     std::string nold_arg = input.getCmdOption("-n");
+    std::string grouping_filename = input.getCmdOption("-g");
+    std::string grouping_columns = input.getCmdOption("-c");
     std::string nsubsteps_arg = input.getCmdOption("--n-substeps");
     std::string gunifrac_arg = input.getCmdOption("-a");
     std::string mode_arg = input.getCmdOption("--mode");
@@ -467,6 +613,10 @@ int main(int argc, char **argv){
     std::string format_arg = input.getCmdOption("--format");
     std::string sformat_arg = input.getCmdOption("-r");
     std::string pcoa_arg = input.getCmdOption("--pcoa");
+    std::string permanova_arg = input.getCmdOption("--permanova");
+    std::string seed_arg = input.getCmdOption("--seed");
+    std::string subsample_depth_arg = input.getCmdOption("--subsample-depth");
+    std::string n_subsamples_arg = input.getCmdOption("--n-subsamples");
     std::string diskbuf_arg = input.getCmdOption("--diskbuf");
 
     if(nsubsteps_arg.empty()) {
@@ -515,13 +665,13 @@ int main(int argc, char **argv){
 
     Format format_val = format_invalid;
     if(!format_arg.empty()) {
-      format_val = get_format(format_arg,method_string);
+      format_val = get_format(format_arg,method_string,mode_arg);
     } else {
-      format_val = get_format(sformat_arg,method_string);
+      format_val = get_format(sformat_arg,method_string,mode_arg);
       format_arg=sformat_arg; // easier to use a single variable
     }
     if(format_val==format_invalid) {
-        err("Invalid format, must be one of ascii|hdf5|hdf5_fp32|hdf5_fp64");
+        err("Invalid format, must be one of ascii|hdf5|hdf5_fp32|hdf5_fp64|hdf5_nodist");
         return EXIT_FAILURE;
     }
 
@@ -531,19 +681,72 @@ int main(int argc, char **argv){
     else
         pcoa_dims = atoi(pcoa_arg.c_str());
 
+    unsigned int permanova_perms;
+    if(permanova_arg.empty()) {
+        if (grouping_filename.empty() || grouping_columns.empty()) {
+            // cannot compute permanova without the grouping file and columns
+            permanova_perms = 0;
+        } else {
+            permanova_perms = 999;
+        }
+    } else {
+        permanova_perms = atoi(permanova_arg.c_str());
+    }
+
+    if(!seed_arg.empty()) {
+         ssu_set_random_seed(atoi(seed_arg.c_str()));
+    }
+
+    unsigned int subsample_depth = 0;
+    if(subsample_depth_arg.empty()) {
+        if(mode_arg == "multi" || mode_arg == "multiple") {
+           err("--subsample-depth required in multi mode.");
+           return EXIT_FAILURE;
+        } else {
+           subsample_depth = 0;
+        }
+    } else {
+        subsample_depth = atoi(subsample_depth_arg.c_str());
+    }
+
+    unsigned int n_subsamples = 0;
+    if(n_subsamples_arg.empty()) {
+        n_subsamples = 100;
+    } else {
+        if(mode_arg == "multi" || mode_arg == "multiple") {
+           n_subsamples = atoi(n_subsamples_arg.c_str());
+        } else {
+           err("--n-subsamples only allowed in multi mode.");
+           return EXIT_FAILURE;
+        }
+    }
 
     if(mode_arg.empty() || mode_arg == "one-off")
-        return mode_one_off(table_filename, tree_filename, output_filename, format_arg, format_val, method_string, pcoa_dims, vaw, g_unifrac_alpha, bypass_tips, nsubsteps, diskbuf_arg);
-    else if(mode_arg == "partial")
+        return mode_one_off(table_filename, tree_filename, output_filename,  format2str(format_val), format_val, method_string,
+                            subsample_depth,
+                            pcoa_dims, permanova_perms, grouping_filename, grouping_columns,
+                            vaw, g_unifrac_alpha, bypass_tips, nsubsteps, diskbuf_arg);
+    else if(mode_arg == "partial") {
+        if (subsample_depth>0) {
+          err("Cannot subsample in partial mode.");
+          return EXIT_FAILURE;
+        }
         return mode_partial(table_filename, tree_filename, output_filename, method_string, vaw, g_unifrac_alpha, bypass_tips, nsubsteps, start_stripe, stop_stripe);
-    else if(mode_arg == "merge-partial")
-        return mode_merge_partial(output_filename, format_val, pcoa_dims, partial_pattern, diskbuf_arg);
+    } else if(mode_arg == "merge-partial")
+        return mode_merge_partial(output_filename, format_val,
+                                  pcoa_dims, permanova_perms, grouping_filename, grouping_columns,
+                                  partial_pattern, diskbuf_arg);
     else if(mode_arg == "check-partial")
         return mode_check_partial(partial_pattern);
     else if(mode_arg == "partial-report")
         return mode_partial_report(table_filename, uint32_t(n_partials), bare);
+    else if(mode_arg == "multi" || mode_arg == "multiple")
+        return mode_multi(table_filename, tree_filename, output_filename, format2str(format_val), format_val, method_string,
+                            n_subsamples,subsample_depth,
+                            pcoa_dims, permanova_perms, grouping_filename, grouping_columns,
+                            vaw, g_unifrac_alpha, bypass_tips, nsubsteps, diskbuf_arg);
     else 
-        err("Unknown mode. Valid options are: one-off, partial, merge-partial, check-partial, partial-report");
+        err("Unknown mode. Valid options are: one-off, partial, merge-partial, check-partial, partial-report, multi");
 
     return EXIT_SUCCESS;
 }
